//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//           Class used for tracking objects for gray scaled images
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//---------------------------------------------------------------------------
#ifndef CTrackH
#define CTrackH
//---------------------------------------------------------------------------
#include "highgui.h"
#include "cv.h"
#include "cvaux.h"
#include <string.h>
#include <math.h>
#include <stdio.h>
#include "cmyimage.h"
#include "CPixel.h"

enum IMAGE_TYPE 	{VARIANCE_IMAGE, FLIR_IMAGE};

class CTrack
{
	// Declaration of the variables
 	private:
    	CMyImage FLIR_image;

        CPixel 	target_center;
        CPixel 	candidate_center;

        int		model_window_radius;
        int		candidate_window_radius;

        int		number_of_bins;
        float	**FLIR_model_color_probabilities;
        float	**FLIR_candidate_color_probabilities;
        float	**FLIR_weights;
	// Declaration of the functions
    public:
    	CTrack();
    	~CTrack();
	    void track_target(char*,char*,char*,char*,int,int,CPixel,int,int);
		

    private:
        float 	kronecker_delta_function			(float,float);

        float	eponechnikov_profile				(float);
        float	derivative_of_eponechnikov_profile	(float);

        float	constant_C 							();		//for model
        float	constant_Ch			  				();		//for candidate

        void	model_density_function_q			();//for model
        void	candidate_density_function_p		();//for candidate

        float	bhattacharyya_coefficient();
        void	calculate_weights( );
        void	determine_new_candidate_location();
        float 	track_target_in_consecutive_frames();
};


#endif
